#define ZASM_LIST_C 1

#include <string.h>
#include <zasm/list.h>
#include <zasm/memory.h>
#include <zasm/errors.h>

#define BLOCK_SIZE 32

struct list_t *list_new()
{
  struct list_t *val;
  
  ALLOCATE(struct list_t, val, 1);
  list_fill(val);
  return val;
}

void list_fill(struct list_t *list)
{
  list->data = NULL;
  list->mem = NULL;
  list->begin = 0;
  list->end = 0;
  list->mem_begin = 0;
  list->mem_end = 0;
}

void list_print(struct list_t *list, FILE *file)
{
  int i;
  
  if (!list)
    {
      fprintf(file, "<list (nil)>");
      return;
    }
  fprintf(file, "<list %i-%i, [", list->begin,list->end);
  for (i = list->begin; i < list->end; i++)
    token_print(list->data[i],file);
  fprintf(file, "]>");
}

void list_reverse(struct list_t *list)
{
  int i;
  struct token_t tmp;

  for (i = 0; i < (list->end - list->begin)/2; i++)
    {
      tmp = list->data[list->begin+i];
      list->data[list->begin+i] = list->data[list->end-i-1];
      list->data[list->end-i-1] = tmp;
    }
}

int list_length(struct list_t *list)
{
  return list->end - list->begin;
}

int list_is_empty(struct list_t *list)
{
  return (list->begin >= list->end);
}

void list_start_begin(struct list_t *list)
{
  list->index = list->begin;
}

void list_start_end(struct list_t *list)
{
  list->index = list->end;
}

struct token_t list_read_begin(struct list_t *list)
{
  struct token_t tok;

  if ((list->index >= list->end)||(list->index < list->begin))
    {
      tok.type = TOK_LIST_END;
      return tok;
    }
  tok = list->data[list->index];
  list->index++;
  return tok;
}

struct token_t list_read_end  (struct list_t *list)
{
  struct token_t tok;

  if ((list->index <= list->begin)||(list->index > list->end))
    {
      tok.type = TOK_LIST_END;
      return tok;
    }
  list->index--;
  tok = list->data[list->index];
  return tok;
}

struct token_t list_read_index(struct list_t *list, int index)
{
  int i;
  struct token_t tok;

  if (index < 0)
    {
      i = list->end + index;
      if (i < list->begin)
	{
	  tok.type = TOK_LIST_END;
	  return tok;
	}
    }
  else
    {
      i = list->begin + index;
      if (i >= list->end)
	{
	  tok.type = TOK_LIST_END;
	  return tok;
	}
    }
  return list->data[i];
}

struct token_t list_get_begin(struct list_t *list)
{
  struct token_t tok;

  if (list->begin >= list->end)
    {
      list->begin = 0;
      list->end = 0;
      tok.type = TOK_LIST_END;
      return tok;
    }
  tok = list->data[list->begin];
  list->begin++;
  return tok;
}

struct token_t list_get_end(struct list_t *list)
{
  struct token_t tok;
  if (list->begin >= list->end)
    {
      list->begin = 0;
      list->end = 0;
      tok.type = TOK_LIST_END;
      return tok;
    }
  list->end--;
  return list->data[list->end];
}

void list_add_begin(struct list_t *list, struct token_t token)
{
  if (list->begin-1 <= list->mem_begin)
    {
      list_allocate(list, list->mem_end - list->mem_begin + BLOCK_SIZE);
      memmove(&(list->data[list->begin+BLOCK_SIZE]), &(list->data[list->begin]),
	      sizeof(struct token_t)*(list->end - list->begin));
      list->mem_begin -= BLOCK_SIZE;
      list->data += BLOCK_SIZE;
    }
  list->begin--;
  list->data[list->begin] = token;
}

void list_add_end(struct list_t *list, struct token_t token)
{
  if (list->end >= list->mem_end)
    {
      list_allocate(list, list->mem_end - list->mem_begin + BLOCK_SIZE);
      list->mem_end += BLOCK_SIZE;
    }
  list->data[list->end] = token;
  list->end++;
}

struct list_t *list_concat(struct list_t *list1, struct list_t *list2)
{
  struct list_t *list;

  list = list_new();
  list_add_list_end(list, list1);
  list_add_list_end(list, list2);
  return list;
}

void list_add_list_begin(struct list_t *list, struct list_t *tokens)
{
  int length, l;
  
  length = tokens->end - tokens->begin;
  if (list->begin - length <= list->mem_begin)
    {
      for (l = 0; list->begin - length <= list->mem_begin - l; l+=BLOCK_SIZE);
      list_allocate(list, list->mem_end - list->mem_begin + l);
      memmove(&(list->data[list->begin+l]), &(list->data[list->begin]),
	      sizeof(struct token_t)*(list->end - list->begin));
      list->mem_begin -= l;
      list->data += l;
    }
  
  list->begin -= length;
  memmove(&(list->data[list->begin]), &(tokens->data[tokens->begin]), length*sizeof(struct token_t));
}

void list_add_list_end(struct list_t *list, struct list_t *tokens)
{
  int length, l;
  
  length = tokens->end - tokens->begin;

  if (list->end + length >= list->mem_end)
    {
      for (l = 0; list->end + length >= list->mem_end + l; l+=BLOCK_SIZE);
      list_allocate(list, list->mem_end - list->mem_begin + l);
      list->mem_end += l;
    }
  
  memmove(&(list->data[list->end]), &(tokens->data[tokens->begin]), length*sizeof(struct token_t));
  list->end += length;
}


void list_allocate(struct list_t *list, int count)
{
  REALLOCATE(struct token_t, (list->mem), count);
  list->data = list->mem - list->mem_begin;
}

void list_set_original(struct list_t *list, struct token_t original)
{
  int i;
  struct token_t *orig_p;

  orig_p = token_make_pointer(original);
  for (i = list->begin; i < list->end; i++)
    {
      list->data[i].position = token_position_set_original(list->data[i].position, token_copy_pointer(orig_p));
    }
  token_release_pointer(orig_p);
}

struct list_t *list_copy(struct list_t *list)
{
  struct list_t *new;
  new = list_new();
  list_add_list_end(new,list);
  list_inc_refcount(new);
  return new;
}

void list_inc_refcount(struct list_t *list)
{
  int i;
  if (!list)
    return;
  for (i = list->begin; i < list->end; i++)
    list->data[i] = token_copy(list->data[i]);
}

void list_dec_refcount(struct list_t *list)
{
  int i;
  if (!list)
    return;
  for (i = list->begin; i < list->end; i++)
    token_free(list->data[i]);
}

void list_delete(struct list_t *list)
{
  if (list->mem)
    free(list->mem);
}

void list_free(struct list_t *list)
{
  list_delete(list);
  free(list);
}
