#define ZASM_TOKENS_C 1

#include <zasm/tokens.h>
#include <zasm/string.h>
#include <zasm/errors.h>
#include <zasm/memory.h>

#define TOKEN_REFCOUNT_T(O) ((struct token_refcount_t *)(O))
struct token_refcount_t
{
  struct token_t token;
  unsigned int refcount;
};

#define TOKEN_ARRAY_SIZE 2048
struct token_refcount_alloc_t
{
  struct token_refcount_t tokens[TOKEN_ARRAY_SIZE];
  struct token_refcount_alloc_t *next;
  struct token_refcount_alloc_t *previous;
};

static struct token_refcount_alloc_t *token_array = NULL;
static unsigned int ta_index;
static unsigned int ta_free[128];
static unsigned int ta_free_index;

static struct token_t *token_make_pointer_search(struct token_refcount_alloc_t *alloc)
{
  int j;

  if (ta_free_index == 0)
    {
      for (j=0; j < TOKEN_ARRAY_SIZE; j++)
	{
	  if (alloc->tokens[j].refcount == 0)
	    {
	      if (ta_free_index >= 128)
		return &(alloc->tokens[j].token);
	      ta_free[ta_free_index] = j;
	      ta_free_index++;
	    }
	}
    }
  if (ta_free_index > 0)
    {
      ta_free_index--;
      return &(alloc->tokens[ta_free[ta_free_index]].token);
    }
  return NULL;
}

static struct token_t *token_make_pointer_get(void)
{
  struct token_refcount_alloc_t *alloc;
  struct token_t *token;

  if (!token_array)
    {
      ALLOCATE(struct token_refcount_alloc_t, token_array, 1);
      ta_index = 0;
      ta_free_index = 0;
      token_array->next = token_array;
      token_array->previous = token_array;
    }
  if (ta_index >= TOKEN_ARRAY_SIZE)
    {
      alloc = token_array;
      do
	{
	  token = token_make_pointer_search(alloc);
	  if (token)
	    {
	      token_array = alloc;
	      return token;
	    }
	  alloc = alloc->next;
	} while (alloc != token_array);
      ALLOCATE(struct token_refcount_alloc_t, alloc, 1);
      alloc->next = token_array;
      alloc->previous = token_array->previous;
      alloc->previous->next = alloc;
      token_array->previous = alloc;
      ta_index = 0;
      ta_free_index = 0;
      token_array = alloc;
    }
  ta_index++;
  return &(token_array->tokens[ta_index-1].token);
}

struct token_t *token_make_pointer(struct token_t token)
{
  struct token_t *tp;
  tp = token_make_pointer_get();
  *tp = token_copy(token);
  TOKEN_REFCOUNT_T(tp)->refcount = 1;
  return tp;
}

struct token_t *token_copy_pointer(struct token_t *token)
{
  TOKEN_REFCOUNT_T(token)->refcount++;
  token_copy(*token);
  return token;
}

void token_release_pointer(struct token_t *token)
{
  if (TOKEN_REFCOUNT_T(token)->refcount == 0)
    ERROR("internal error\n");
  TOKEN_REFCOUNT_T(token)->refcount--;
  token_free(*token);
}

struct position_t token_position_fill(char *file, int line_begin, int line_end, struct token_t *token)
{
  struct position_t pos;
  pos.file = file;
  pos.line_begin = line_begin;
  pos.line_end = line_end;
  pos.original = token;
  return pos;
}

struct position_t token_position_copy(struct position_t pos)
{
  if (pos.file)
    string_copy(pos.file);
  if (pos.original)
    token_copy_pointer(pos.original);
  return pos;
}

struct position_t token_position_set_original(struct position_t pos, struct token_t *original)
{
  if (!pos.original)
    {
      pos.original = original;
      return pos;
    }
  pos.original = token_make_pointer(*(pos.original));
  pos.original->position = token_position_set_original(pos.original->position, original);
  return pos;
}

struct position_t token_position_set_original_longest(struct position_t pos, struct token_t o1, struct token_t o2)
{
  int l1, l2;
  struct token_t *t;

  for (l1 = 0, t = o1.position.original; t; t = t->position.original)
    l1++;
  for (l2 = 0, t = o2.position.original; t; t = t->position.original)
    l2++;
  if (l1 < l2)
    t = token_make_pointer(token_copy(o2));
  else
    t = token_make_pointer(token_copy(o1));
  return token_position_set_original(pos, t);
}

void token_position_free(struct position_t pos)
{
  if (pos.file)
    string_free(pos.file);
  if (pos.original)
    token_release_pointer(pos.original);
}

struct token_t token_copy(struct token_t token)
{
  switch (token.type)
    {
      // will be auto-gen
case TOK_STRING:
case TOK_ISTRING:
case TOK_NAME:
case TOK_VARARG:
case TOK_END_FILE:

      token.value.t_char_p = string_copy(token.value.t_char_p);
      break;
    default:
      break;
    }
  token_position_copy(token.position);
  return token;
}

int token_value_neutral_check(int value, int mask)
{
  if (~mask == 0)
    return 1;
  return !((value > mask) || (value < (-(mask>>1)) - 1));
}

int token_value_signed_check(int value, int mask)
{
  mask>>=1;
  return !((value > mask) || (value < (-mask) - 1));
}

int token_value_unsigned_check(int value, int mask)
{
  return !(value&~mask);
}

void token_print(struct token_t token, FILE *file)
{
#define WRITE(FMT,A...) fprintf(file, FMT, ## A)
#define CHECK_NEUTRAL(VAL,MASK) (({if (!token_value_neutral_check(VAL,MASK)) T_WARNING(token,"value out of bounds\n");}))
#define CHECK_SIGNED(VAL,MASK) (({if (!token_value_signed_check(VAL,MASK)) T_WARNING(token,"value out of bounds\n");}))
#define CHECK_UNSIGNED(VAL,MASK) (({if (!token_value_unsigned_check(VAL,MASK)) T_WARNING(token,"value out of bounds\n");}))
  switch (token.type)
    {
case TOK_MULT_LIST: break;
case TOK_PROCESSOR: break;
case TOK_PROC_INSTR: break;
case TOK_PROC_INSTR_ARG: break;
case TOK_LITTLE_ENDIAN: {WRITE("LITTLE ENDIAN");}break;
case TOK_BIG_ENDIAN: {WRITE("BIG ENDIAN");}break;
case TOK_ADDINSTR: {WRITE("#addinstr ");}break;
case TOK_BYTE_INT: {CHECK_NEUTRAL(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_WORD_INT: {CHECK_NEUTRAL(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_DWORD_INT: {CHECK_NEUTRAL(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_BYTE_INT_P: {CHECK_NEUTRAL(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_WORD_INT_P: {CHECK_NEUTRAL(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_DWORD_INT_P: {CHECK_NEUTRAL(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_BYTE_INT: {CHECK_SIGNED(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_SIGNED_WORD_INT: {CHECK_SIGNED(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_SIGNED_DWORD_INT: {CHECK_SIGNED(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_SIGNED_BYTE_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_WORD_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_DWORD_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_UNSIGNED_BYTE_INT: {CHECK_UNSIGNED(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_UNSIGNED_WORD_INT: {CHECK_UNSIGNED(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_UNSIGNED_DWORD_INT: {CHECK_UNSIGNED(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_UNSIGNED_BYTE_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_UNSIGNED_WORD_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_UNSIGNED_DWORD_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_BYTE: {WRITE("#byte ");}break;
case TOK_WORD: {WRITE("#word ");}break;
case TOK_DWORD: {WRITE("#dword ");}break;
case TOK_SIGNED_BYTE: {WRITE("#sbyte ");}break;
case TOK_SIGNED_WORD: {WRITE("#sword ");}break;
case TOK_SIGNED_DWORD: {WRITE("#sdword ");}break;
case TOK_UNSIGNED_BYTE: {WRITE("#sbyte ");}break;
case TOK_UNSIGNED_WORD: {WRITE("#sword ");}break;
case TOK_UNSIGNED_DWORD: {WRITE("#sdword ");}break;
case TOK_ARG: {WRITE("@%i", token.value.t_int);}break;
case TOK_DB: {WRITE(".db ");}break;
case TOK_DW: {WRITE(".dw ");}break;
case TOK_DD: {WRITE(".dd ");}break;
case TOK_NONE: break;
case TOK_LIST_END: break;
case TOK_STREAM_END: break;
case TOK_END_OF_FILE: break;
case TOK_STRING: {WRITE("\"%s\"", token.value.t_char_p);}break;
case TOK_ISTRING: {WRITE("<%s>", token.value.t_char_p);}break;
case TOK_NAME: {WRITE("%s", token.value.t_char_p);}break;
case TOK_VARARG: {WRITE("%s...", token.value.t_char_p);}break;
case TOK_NEWLINE: {WRITE("\n");}break;
case TOK_ESCAPED_NEWLINE: {WRITE("\n");}break;
case TOK_SPACE: {WRITE(" ");}break;
case TOK_SPECIAL: {WRITE("%c", token.value.t_char);}break;
case TOK_INTEGER: {WRITE("%i", token.value.t_int);}break;
case TOK_BEGIN_INSTR: {WRITE("#bi ");}break;
case TOK_ALIGN_ADDR: {WRITE("#align_addr ");}break;
case TOK_ALIGN_MULT: {WRITE("#align_mult ");}break;
case TOK_ORG: {WRITE("#org ");}break;
case TOK_PAGE: {WRITE("#setpage ");}break;
case TOK_PAGE_START: {WRITE("#page_start ");}break;
case TOK_NEWPAGE: {WRITE("#newpage ");}break;
case TOK_EQU: {WRITE("#equ ");}break;
case TOK_EXPORT: {WRITE("#export ");}break;
case TOK_BEGIN_SECTION: {WRITE("#begin_section ");}break;
case TOK_END_SECTION: {WRITE("#end_section ");}break;
case TOK_SHARP: {WRITE("#");}break;
case TOK_BLOCK: {WRITE("#block ");}break;
case TOK_FILL: {WRITE("#fill ");}break;
case TOK_SET_PAGE: break;
case TOK_PC_FAIL: {WRITE("<pc fail>");}break;
case TOK_PC_FAIL1: {WRITE("<pc fail1>");}break;
case TOK_PC_SUCCESS: {WRITE("<pc success>");}break;
case TOK_BACKSLASH: {WRITE("\\");}break;
case TOK_EXPR: {WRITE("<expr>");}break;
case TOK_LIST: {WRITE("<list>");}break;
case TOK_END_FILE: {WRITE("<endFile>");}break;
case TOK_OP_IFTHEN: {WRITE("?");}break;
case TOK_OP_ELSE: {WRITE(":");}break;
case TOK_OP_LOG_OR: {WRITE("||");}break;
case TOK_OP_LOG_AND: {WRITE("&&");}break;
case TOK_OP_LOG_XOR: {WRITE("^^");}break;
case TOK_OP_LOG_NOT: {WRITE("!");}break;
case TOK_OP_BIN_OR: {WRITE(".|");}break;
case TOK_OP_BIN_AND: {WRITE(".&");}break;
case TOK_OP_BIN_XOR: {WRITE(".^");}break;
case TOK_OP_BIN_NOT: {WRITE(".~");}break;
case TOK_OP_SHIFT_LEFT: {WRITE("<<");}break;
case TOK_OP_SHIFT_RIGHT: {WRITE(">>");}break;
case TOK_OP_EQU: {WRITE("==");}break;
case TOK_OP_NEQ: {WRITE("!=");}break;
case TOK_OP_LT: {WRITE("<");}break;
case TOK_OP_GT: {WRITE(">");}break;
case TOK_OP_LE: {WRITE("<=");}break;
case TOK_OP_GE: {WRITE(">=");}break;
case TOK_OP_CMP: {WRITE("<=>");}break;
case TOK_OP_ADD: {WRITE("+");}break;
case TOK_OP_SUB: {WRITE("-");}break;
case TOK_OP_MUL: {WRITE("*");}break;
case TOK_OP_DIV: {WRITE("/");}break;
case TOK_OP_MOD: {WRITE("%%");}break;
case TOK_OP_POW: {WRITE("**");}break;
case TOK_DOT: {WRITE(".");}break;
case TOK_PAREN_OPEN: {WRITE("(");}break;
case TOK_PAREN_CLOSE: {WRITE(")");}break;
case TOK_BRACE_OPEN: {WRITE("{");}break;
case TOK_BRACE_CLOSE: {WRITE("}");}break;
case TOK_SQUARE_OPEN: {WRITE("[");}break;
case TOK_SQUARE_CLOSE: {WRITE("]");}break;
case TOK_COMMA: {WRITE(", ");}break;
case TOK_PASTE: {WRITE(" ## ");}break;
case TOK_DEFINED: {WRITE("defined");}break;
case TOK_DEFINE: {WRITE("#define");}break;
case TOK_MACRO: {WRITE(".macro");}break;
case TOK_REDEF: {WRITE("#redef");}break;
case TOK_UNDEF: {WRITE("#undef");}break;
case TOK_ENDM: {WRITE("#endm");}break;
case TOK_PP_BEGIN_SECTION: {WRITE("#pp_begin_section");}break;
case TOK_PP_END_SECTION: {WRITE("#pp_end_section");}break;
case TOK_INCLUDE: {WRITE("#include");}break;
case TOK_IF: {WRITE("#if");}break;
case TOK_THEN: {WRITE("#then");}break;
case TOK_IFDEF: {WRITE("#ifdef");}break;
case TOK_IFNDEF: {WRITE("#ifndef");}break;
case TOK_ELSE: {WRITE("#else");}break;
case TOK_ELSEIF: {WRITE("#elseif");}break;
case TOK_ELSEIFDEF: {WRITE("#elseifdef");}break;
case TOK_ELSEIFNDEF: {WRITE("#elseifndef");}break;
case TOK_ENDIF: {WRITE("#endif");}break;
case TOK_FOR: {WRITE("#for");}break;
case TOK_IN: {WRITE("#in");}break;
case TOK_FROM: {WRITE("#from");}break;
case TOK_TO: {WRITE("#to");}break;
case TOK_STEP: {WRITE("#step");}break;
case TOK_DO: {WRITE("#do");}break;
case TOK_ENDFOR: {WRITE("#endfor");}break;
case TOK_ERROR: {WRITE("#error");}break;
case TOK_WARNING: {WRITE("#warning");}break;
case TOK_DEBUG: {WRITE("#debug");}break;
case TOK_FEED_ACTION: break;
case TOK_FEED_LIST: break;
case TOK_OUTPUT: {WRITE("#output ");}break;

    default:
      ERROR("unknown token\n");
      break;
    }
  return;
#undef WRITE
#undef CHECK_NEUTRAL
#undef CHECK_SIGNED
#undef CHECK_UNSIGNED
}

void token_print_nowarning(struct token_t token, FILE *file)
{
#define WRITE(FMT,A...) fprintf(file, FMT, ## A)
#define CHECK_NEUTRAL(VAL,MASK) (({}))
#define CHECK_SIGNED(VAL,MASK) (({}))
#define CHECK_UNSIGNED(VAL,MASK) (({}))
  switch (token.type)
    {
case TOK_MULT_LIST: break;
case TOK_PROCESSOR: break;
case TOK_PROC_INSTR: break;
case TOK_PROC_INSTR_ARG: break;
case TOK_LITTLE_ENDIAN: {WRITE("LITTLE ENDIAN");}break;
case TOK_BIG_ENDIAN: {WRITE("BIG ENDIAN");}break;
case TOK_ADDINSTR: {WRITE("#addinstr ");}break;
case TOK_BYTE_INT: {CHECK_NEUTRAL(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_WORD_INT: {CHECK_NEUTRAL(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_DWORD_INT: {CHECK_NEUTRAL(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_BYTE_INT_P: {CHECK_NEUTRAL(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_WORD_INT_P: {CHECK_NEUTRAL(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_DWORD_INT_P: {CHECK_NEUTRAL(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_BYTE_INT: {CHECK_SIGNED(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_SIGNED_WORD_INT: {CHECK_SIGNED(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_SIGNED_DWORD_INT: {CHECK_SIGNED(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_SIGNED_BYTE_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_WORD_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_DWORD_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_UNSIGNED_BYTE_INT: {CHECK_UNSIGNED(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_UNSIGNED_WORD_INT: {CHECK_UNSIGNED(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_UNSIGNED_DWORD_INT: {CHECK_UNSIGNED(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_UNSIGNED_BYTE_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_UNSIGNED_WORD_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_UNSIGNED_DWORD_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_BYTE: {WRITE("#byte ");}break;
case TOK_WORD: {WRITE("#word ");}break;
case TOK_DWORD: {WRITE("#dword ");}break;
case TOK_SIGNED_BYTE: {WRITE("#sbyte ");}break;
case TOK_SIGNED_WORD: {WRITE("#sword ");}break;
case TOK_SIGNED_DWORD: {WRITE("#sdword ");}break;
case TOK_UNSIGNED_BYTE: {WRITE("#sbyte ");}break;
case TOK_UNSIGNED_WORD: {WRITE("#sword ");}break;
case TOK_UNSIGNED_DWORD: {WRITE("#sdword ");}break;
case TOK_ARG: {WRITE("@%i", token.value.t_int);}break;
case TOK_DB: {WRITE(".db ");}break;
case TOK_DW: {WRITE(".dw ");}break;
case TOK_DD: {WRITE(".dd ");}break;
case TOK_NONE: break;
case TOK_LIST_END: break;
case TOK_STREAM_END: break;
case TOK_END_OF_FILE: break;
case TOK_STRING: {WRITE("\"%s\"", token.value.t_char_p);}break;
case TOK_ISTRING: {WRITE("<%s>", token.value.t_char_p);}break;
case TOK_NAME: {WRITE("%s", token.value.t_char_p);}break;
case TOK_VARARG: {WRITE("%s...", token.value.t_char_p);}break;
case TOK_NEWLINE: {WRITE("\n");}break;
case TOK_ESCAPED_NEWLINE: {WRITE("\n");}break;
case TOK_SPACE: {WRITE(" ");}break;
case TOK_SPECIAL: {WRITE("%c", token.value.t_char);}break;
case TOK_INTEGER: {WRITE("%i", token.value.t_int);}break;
case TOK_BEGIN_INSTR: {WRITE("#bi ");}break;
case TOK_ALIGN_ADDR: {WRITE("#align_addr ");}break;
case TOK_ALIGN_MULT: {WRITE("#align_mult ");}break;
case TOK_ORG: {WRITE("#org ");}break;
case TOK_PAGE: {WRITE("#setpage ");}break;
case TOK_PAGE_START: {WRITE("#page_start ");}break;
case TOK_NEWPAGE: {WRITE("#newpage ");}break;
case TOK_EQU: {WRITE("#equ ");}break;
case TOK_EXPORT: {WRITE("#export ");}break;
case TOK_BEGIN_SECTION: {WRITE("#begin_section ");}break;
case TOK_END_SECTION: {WRITE("#end_section ");}break;
case TOK_SHARP: {WRITE("#");}break;
case TOK_BLOCK: {WRITE("#block ");}break;
case TOK_FILL: {WRITE("#fill ");}break;
case TOK_SET_PAGE: break;
case TOK_PC_FAIL: {WRITE("<pc fail>");}break;
case TOK_PC_FAIL1: {WRITE("<pc fail1>");}break;
case TOK_PC_SUCCESS: {WRITE("<pc success>");}break;
case TOK_BACKSLASH: {WRITE("\\");}break;
case TOK_EXPR: {WRITE("<expr>");}break;
case TOK_LIST: {WRITE("<list>");}break;
case TOK_END_FILE: {WRITE("<endFile>");}break;
case TOK_OP_IFTHEN: {WRITE("?");}break;
case TOK_OP_ELSE: {WRITE(":");}break;
case TOK_OP_LOG_OR: {WRITE("||");}break;
case TOK_OP_LOG_AND: {WRITE("&&");}break;
case TOK_OP_LOG_XOR: {WRITE("^^");}break;
case TOK_OP_LOG_NOT: {WRITE("!");}break;
case TOK_OP_BIN_OR: {WRITE(".|");}break;
case TOK_OP_BIN_AND: {WRITE(".&");}break;
case TOK_OP_BIN_XOR: {WRITE(".^");}break;
case TOK_OP_BIN_NOT: {WRITE(".~");}break;
case TOK_OP_SHIFT_LEFT: {WRITE("<<");}break;
case TOK_OP_SHIFT_RIGHT: {WRITE(">>");}break;
case TOK_OP_EQU: {WRITE("==");}break;
case TOK_OP_NEQ: {WRITE("!=");}break;
case TOK_OP_LT: {WRITE("<");}break;
case TOK_OP_GT: {WRITE(">");}break;
case TOK_OP_LE: {WRITE("<=");}break;
case TOK_OP_GE: {WRITE(">=");}break;
case TOK_OP_CMP: {WRITE("<=>");}break;
case TOK_OP_ADD: {WRITE("+");}break;
case TOK_OP_SUB: {WRITE("-");}break;
case TOK_OP_MUL: {WRITE("*");}break;
case TOK_OP_DIV: {WRITE("/");}break;
case TOK_OP_MOD: {WRITE("%%");}break;
case TOK_OP_POW: {WRITE("**");}break;
case TOK_DOT: {WRITE(".");}break;
case TOK_PAREN_OPEN: {WRITE("(");}break;
case TOK_PAREN_CLOSE: {WRITE(")");}break;
case TOK_BRACE_OPEN: {WRITE("{");}break;
case TOK_BRACE_CLOSE: {WRITE("}");}break;
case TOK_SQUARE_OPEN: {WRITE("[");}break;
case TOK_SQUARE_CLOSE: {WRITE("]");}break;
case TOK_COMMA: {WRITE(", ");}break;
case TOK_PASTE: {WRITE(" ## ");}break;
case TOK_DEFINED: {WRITE("defined");}break;
case TOK_DEFINE: {WRITE("#define");}break;
case TOK_MACRO: {WRITE(".macro");}break;
case TOK_REDEF: {WRITE("#redef");}break;
case TOK_UNDEF: {WRITE("#undef");}break;
case TOK_ENDM: {WRITE("#endm");}break;
case TOK_PP_BEGIN_SECTION: {WRITE("#pp_begin_section");}break;
case TOK_PP_END_SECTION: {WRITE("#pp_end_section");}break;
case TOK_INCLUDE: {WRITE("#include");}break;
case TOK_IF: {WRITE("#if");}break;
case TOK_THEN: {WRITE("#then");}break;
case TOK_IFDEF: {WRITE("#ifdef");}break;
case TOK_IFNDEF: {WRITE("#ifndef");}break;
case TOK_ELSE: {WRITE("#else");}break;
case TOK_ELSEIF: {WRITE("#elseif");}break;
case TOK_ELSEIFDEF: {WRITE("#elseifdef");}break;
case TOK_ELSEIFNDEF: {WRITE("#elseifndef");}break;
case TOK_ENDIF: {WRITE("#endif");}break;
case TOK_FOR: {WRITE("#for");}break;
case TOK_IN: {WRITE("#in");}break;
case TOK_FROM: {WRITE("#from");}break;
case TOK_TO: {WRITE("#to");}break;
case TOK_STEP: {WRITE("#step");}break;
case TOK_DO: {WRITE("#do");}break;
case TOK_ENDFOR: {WRITE("#endfor");}break;
case TOK_ERROR: {WRITE("#error");}break;
case TOK_WARNING: {WRITE("#warning");}break;
case TOK_DEBUG: {WRITE("#debug");}break;
case TOK_FEED_ACTION: break;
case TOK_FEED_LIST: break;
case TOK_OUTPUT: {WRITE("#output ");}break;

    default:
      ERROR("unknown token\n");
      break;
    }
  return;
#undef WRITE
#undef CHECK_NEUTRAL
#undef CHECK_SIGNED
#undef CHECK_UNSIGNED
}

void token_free(struct token_t token)
{
  switch (token.type)
    {
      //will be auto-gen
case TOK_STRING:
case TOK_ISTRING:
case TOK_NAME:
case TOK_VARARG:
case TOK_END_FILE:

      string_free(token.value.t_char_p);
      break;
    default:
      break;
    }
  token_position_free(token.position);
}
